#include <cstdlib>
#include <iostream>
#include <cmath>
#include <ctime>
#include "Vector.h"
#include "Line.h"
#include "Transform.h"
#include "Rectangle.h"

#include "CollisionDetector.h"

#ifndef PI
#define PI 3.14159265358979323846264
#endif

//This is the only function that works exclusively on rectangles
//To make it apply to all polygons, the angle comparisons could be 
//omitted, they just allow faster calculations.
bool CollisionDetector::Test(Rectangle const &boxA, Transform const &trA,
	      Rectangle const &boxB, Transform const &trB)
{
	//Prelim check
	//Find the left-most vertex //rect only optimized //hackshackshacks
	Rectangle a = boxA;
	Rectangle b = boxB;
	int leftMostA = 0;
	int leftMostB = 0;
	
	for(int i = 0; i < 4; ++i)
	{
		a.vertices[i] += trA.pos;
		b.vertices[i] += trB.pos;
		if(a.vertices[i].x < a.vertices[leftMostA].x)
			leftMostA = i;
		if(b.vertices[i].x < b.vertices[leftMostB].x)
			leftMostB = i;
	}

	float Ax = a.vertices[ leftMostA         ].x;
	float Ay = a.vertices[(leftMostA + 1) % 4].y; //top
	float Aw = a.vertices[(leftMostA + 2) % 4].x; //right
	float Ah = a.vertices[(leftMostA + 3) % 4].y; //bottom
	
	float Bx = b.vertices[ leftMostB         ].x;
	float By = b.vertices[(leftMostB + 1) % 4].y; //top
	float Bw = b.vertices[(leftMostB + 2) % 4].x; //right
	float Bh = b.vertices[(leftMostB + 3) % 4].y; //bottom
	
	if(Ax >= Bw) return false;
	if(Aw <= Bx) return false;
	if(Ay >= Bh) return false;
	if(Ah <= By) return false;

	//If the angle is zero, no further tests required
	if(trA.rot == 0 && trB.rot == 0) return true;

	//If the angles are the same, rotate to zero and repeat
	/*else if(transformA.rot == transformB.rot)
	{
		Rectangle a = boxA;
		Rectangle b = boxB;
		Transform tA = transformA;
		Transform tB = transformB;
		tA.rot = 0;
		tB.rot = 0;

		for (int i = 0; i < 4; i++)
	       	{
		*	a.vertices[i] = Rotate(a.vertices[i], transformA.rot);
			b.vertices[i] = Rotate(b.vertices[i], transformB.rot);
		
			a.vertices[i] += transformA.pos;
			b.vertices[i] += transformB.pos;
		*
			tA.pos = Rotate(tA.pos, -transformA.rot);
			tB.pos = Rotate(tB.pos, -transformB.rot);
			a.vertices[i] = Rotate(a.vertices[i], -transformA.rot);
			b.vertices[i] = Rotate(b.vertices[i], -transformB.rot);
		}

		return CollisionTest(a,tA,b,tB);
	}*/
	//Otherwise do the full calculation
	
	for(int i = 0; i < 4; ++i)
	{
		Line aLine(a.vertices[i],a.vertices[ (i+1) % 4 ]);
		for(int j = 0; j < 4; ++j)
		{
			Line bLine(b.vertices[j],b.vertices[ (j+1) % 4 ]);

			if(Intersect(aLine,bLine))
				return true;
		}
	}

	return false;
}

//Work out whether two lines intersect
bool CollisionDetector::Intersect(Line const &line1, Line const &line2)
{
	//If the gradients are almost the same, save ourselves 
	//some unnecessary calculation by assuming no collision
	if (fabs(line1.dydx() - line2.dydx()) < 0.01)
	{
		return false;
	}

	//If there is no change in either coordinate, we
	//have a point. (This should never happen).
	if( (line1.dx() == 0 && line1.dy() == 0) ||
	    (line2.dx() == 0 && line2.dy() == 0) )
	{
		return false;
	}
	//This is a special case caused by infinities
	//where one line is vertical and one horizontal.
	if(line1.dy() == 0 && line2.dx() == 0)
	{
		//Is arg1 between arg2 and arg3?
		if(Between(line1.vertex1.y,line2.vertex1.y,line2.vertex2.y))
			return Between(line2.vertex1.x,line1.vertex1.x,line1.vertex2.x);
		//Co-ords = (line1.vertex1.x,line2.vertex1.y)
	}
	else if(line2.dy() == 0 && line1.dx() == 0)
	{
		if(Between(line2.vertex1.y,line1.vertex1.y, line1.vertex2.y))
			return Between(line1.vertex1.x,line2.vertex1.x,line2.vertex2.x);
		//Co-ords = (line2.vertex1.x,line1.vertex1.y)
	}

	//When all else fails...
	if(Intersect2(line1,line2))
	{
		return Intersect2(line2,line1);
	}

	return false;
}

bool CollisionDetector::Intersect2(Line const& line1, Line const& line2)
{
	//The lines can be thought of as a point plus
	//a direction multiplied by some unknown factor.
	//The result is calculated by setting the two
	//lines equal to each other, calculating the
	//vector product of each side with the direction
	//of l2, and solving for the unknown factor.
	//
	//result is the length along line1 at which the
	//lines intersect. If it is less than 0 or greater
	//than the length of line1, then the line segments
	//do not meet

	float result;

	if(line1.dx() == 0)
	{
		//If dx is 0, dy/dx is infinity, and
		//we have to use dx/dy

		//std::cout << "dx/dy" << std::endl;
		
		result = ( line2.dxdy()
			 * ( line2.vertex1.y - line1.vertex1.y )
			 - ( line2.vertex1.x - line1.vertex1.x )
			 ) / ( line1.dxdy() - line2.dxdy() );

		//std::cout << result << " vs. " << line1.dy() << std::endl;

		if(line1.dy() > 0)
		{
			return result <= line1.dy() && result > 0;
		}
		else
		{
			return result >= line1.dy() && result < 0;
		}

		//Co-ords = (result*(line1.dxdy), result)
	}
	else
	{
		//std::cout << "dy/dx" << std::endl;
		result = ( line2.dydx() 
			 * ( -line1.vertex1.x + line2.vertex1.x )
			 - ( -line1.vertex1.y + line2.vertex1.y )
			 ) / (line2.dydx() - line1.dydx());

		//std::cout << result << " vs. " << line1.dy() << std::endl;

		if(line1.dx() > 0)
		{
			return result <= line1.dx() && result > 0;
		}
		else
		{
			return result >= line1.dx() && result < 0;
		}

		//Co-ords = (result, result*(line1.dydx))
	}
}

//Work out whether a point lies between two other points
bool CollisionDetector::Between(float p, float q1, float q2)
{
	if(q1 > q2)
	{
		return p <= q1 && p >= q2;
	}
	else if(q2 > q1)
	{
		return p <= q2 && p >= q1;
	}
	else//q1 == q2
		return p == q1;
}
