#include <SDL/SDL.h>
#include <SDL/SDL_image.h>
#include <SDL/SDL_ttf.h>
#include <GL/gl.h>
#include <iostream>
#include <string>
#include <map>
#include "Texture.h"

#include "Graphics.h"

bool Graphics::_loaded = false;
std::map<GFX,Texture*> Graphics::_gfx;
Texture* Graphics::_fonts[FONT_MAX][128] = {NULL};

bool Graphics::LoadGraphics()
{
	if(!Graphics::_loaded)
	{
		std::cout << "Loading graphics" << std::endl;

		_gfx[GFX_BIT] = CreateTexture(LoadImage("bit.png"));
		_gfx[GFX_GWB] = CreateTexture(LoadImage("gwb.png"));
		_gfx[GFX_MAP] = CreateTexture(LoadImage("map4.png"));
		_gfx[GFX_UNICORN] = CreateTexture(LoadImage("unicorn.png"));

		std::map<GFX,Texture*>::iterator iter;
		for(iter = _gfx.begin(); iter != _gfx.end(); iter++)
		{
			if((*iter).second == NULL)
			{
				std::cout << "Error: Texture " << (*iter).first << " not loaded" << std::endl;
				//Need to free loaded graphics here
				return false;
			}
		}

		if(!LoadTTFFont("libsans.ttf"))
			return false;

		_loaded = true;
	}
	return true;
}

Texture const* Graphics::GetChar(int font,char c)
{
	return _fonts[font][c];
}

Texture const* Graphics::GetTexture(GFX id)
{
	return _gfx[id];
}

SDL_Surface* Graphics::GenerateCharacter(char c,TTF_Font *const font)
{
	SDL_Color colour = {0xFF,0xFF,0xFF};
	SDL_Surface *character = TTF_RenderText_Blended(font,&c,colour);
	if(character == NULL)
	{
		std::cout << "Failed to generate character " << c;
	       	std::cout << "(" << TTF_GetError() << ")" << std::endl;

		return NULL;
	}

	return character;
}

SDL_Surface* Graphics::LoadImage(std::string file)
{
	//Load image into SDL_Surface
	SDL_Surface *image = NULL;
	std::string filepath = "res/gfx/" + file;
	image = IMG_Load(filepath.c_str());
	if(image == NULL)
	{
		std::cout << "Error: Failed to load " << filepath;
		std::cout << "(" << IMG_GetError() << ")" << std::endl;
		
		return NULL;
	}
	

	return image;
}

bool Graphics::LoadTTFFont(std::string file)
{
	TTF_Font *font = NULL;
	std::string filepath = "res/fnt/" + file;
	font = TTF_OpenFont(filepath.c_str(),40);
	if(font == NULL)
	{
		std::cout << "Error: Could not open font (";
		std::cout << TTF_GetError() << ")" << std::endl;

		return false;
	}

	LoadFontTextures(FONT_ONE,font);
	TTF_CloseFont(font);

	return true;
}

void Graphics::LoadFontTextures(Font f,TTF_Font *const font)
{
	SDL_Surface *temp = NULL;
	for(int j = 0; j < 128; j++)
	{
		temp = GenerateCharacter(j,font);
		if (temp != NULL)
		{
			_fonts[f][j] = CreateTexture(temp);
			if(_fonts[f][j] == NULL)
			{
				std::cout << "Error: Could not load letter " << j << std::endl;
			}
		}
	}
}

Texture* Graphics::CreateTexture(SDL_Surface *image, int frames)
{
	//Process surface and check for errors.
	if(image == NULL)
	{
		return NULL;
	}
	if(SDL_SetAlpha(image,0,0xFF) < 0)
	{
		std::cout << "Error: Could not set surface alpha properties ";
		std::cout << "(" << SDL_GetError() << ")" << std::endl;

		SDL_FreeSurface(image);
		return NULL;
	}

	Texture *tex = new Texture();

	const int bpp = image->format->BytesPerPixel;
	const int p2w = NextPowerOfTwo(image->w);
	const int p2h = NextPowerOfTwo(image->h);

	tex->w = image->w;
	tex->h = image->h;
	tex->su = (float)tex->w/p2w;
	tex->sv = (float)tex->h/p2h;
	tex->frames = frames;

	int format;

	switch(bpp)
	{
		case 3:
			tex->alpha = false;
			format = GL_RGB;
			break;
		case 4: 
			tex->alpha = true;
			format = GL_RGBA;
			break;
		default:
			std::cout << "Error: Corrupt image" << std::endl;
			SDL_FreeSurface(image);
			delete(tex);
			return NULL;
			break;
	}

	SDL_Surface *optimized;

	//Convert surface into GL compatible format
	if(p2w != image->w || p2h != image->h)
	{
		optimized = SDL_CreateRGBSurface(
				SDL_SWSURFACE,
				p2w, p2h,
				bpp*8,
				image->format->Rmask,
				image->format->Gmask,
				image->format->Bmask,
				image->format->Amask);

		SDL_BlitSurface(image,NULL,optimized,NULL);
		SDL_FreeSurface(image);
	}
	else
	{
		optimized = image;
	}

	//Load surface as GL Texture
	glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	glGenTextures(1,&tex->id);

	glBindTexture(GL_TEXTURE_2D,tex->id);

	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

	glTexImage2D(GL_TEXTURE_2D,
			0,
			bpp,
			optimized->w,
			optimized->h,
			0,
			format,
			GL_UNSIGNED_BYTE,
			optimized->pixels);

	//Return texture
	SDL_FreeSurface(optimized);

	return tex;
}

inline int Graphics::NextPowerOfTwo(int a)
{
	int x = 1;
	while (x < a)
		x <<= 1;
	return x;
}

void Graphics::UnloadGraphics()
{
	std::cout << "Unloading graphics" << std::endl;

	std::cout << "Deleting textures" << std::endl;
	std::map<GFX,Texture*>::iterator iter;
	for (iter = _gfx.begin(); iter != _gfx.end(); iter++)
	{
		if(iter->second != NULL)
			glDeleteTextures(1,&(iter->second->id));
	}
	std::cout << "Deleting font textures" << std::endl;
	for (int i = 0; i < FONT_MAX; i++)
	{
		for (int j = 0; j < 128; j++)
		{
			if(_fonts[i][j] != NULL)
			{
				glDeleteTextures(1,&_fonts[i][j]->id);
			}
		}
	}

	_loaded = false;

	std::cout << "Graphics unloaded" << std::endl;
}
